from typing import TYPE_CHECKING, Any, Callable, TypeVar, overload

from chorey._types import Feeder, FirstInput, Input, Output
from chorey.utils import get_step_func_name

if TYPE_CHECKING:
    from chorey._types import StepFunc
    from chorey.branch import Branch
    from chorey.route import Route
    from chorey.step import Step

_A = TypeVar("_A")
_B = TypeVar("_B")
_C = TypeVar("_C")
_D = TypeVar("_D")
_E = TypeVar("_E")
_T = TypeVar("_T")
_R = TypeVar("_R")


class BranchableMixin(Feeder[FirstInput, Input, Output]):
    @overload
    def branch(
        self,
        a: "Step[Output, Any, _A]",
    ) -> "Branch[FirstInput, Output, _A]": ...

    @overload
    def branch(
        self,
        a: "Step[Output, Any, _A]",
        b: "Step[Output, Any, _B]",
    ) -> "Branch[FirstInput, Output, _A, _B]": ...

    @overload
    def branch(
        self,
        a: "Step[Output, Any, _A]",
        b: "Step[Output, Any, _B]",
        c: "Step[Output, Any, _C]",
    ) -> "Branch[FirstInput, Output, _A, _B, _C]": ...

    @overload
    def branch(
        self,
        a: "Step[Output, Any, _A]",
        b: "Step[Output, Any, _B]",
        c: "Step[Output, Any, _C]",
        d: "Step[Output, Any, _D]",
    ) -> "Branch[FirstInput, Output, _A, _B, _C, _D]": ...

    @overload
    def branch(
        self,
        a: "Step[Output, Any, _A]",
        b: "Step[Output, Any, _B]",
        c: "Step[Output, Any, _C]",
        d: "Step[Output, Any, _D]",
        e: "Step[Output, Any, _E]",
    ) -> "Branch[FirstInput, Output, _A, _B, _C, _D, _E]": ...

    def branch(self, *args: "Step"):  # type: ignore
        from chorey.branch import Branch

        return Branch(steps=list(args), previous=self)


class RoutableMixin(Feeder[FirstInput, Input, Output]):
    def route(
        self,
        *choices: "Step[Output, Any, _T]",
        selector: Callable[[Output], int],
        decision_label: str | None = None,
    ) -> "Route[FirstInput, Output, _T]":
        from chorey.route import Route

        return Route(
            choices=list(choices),
            selector=selector,
            decision_label=decision_label,
            previous=self,  # type: ignore
        )


class NextableMixin(Feeder[FirstInput, Input, Output]):
    def next(
        self,
        func: "StepFunc[Output, _R]",
        /,
        *,
        description: str | None = None,
        name: str | None = None,
    ) -> "Step[FirstInput, Output, _R]":
        from chorey.step import Step

        return Step[FirstInput, Output, _R](
            func=func,
            name=name or get_step_func_name(func),
            description=description,
            previous=self,
        )


class GetFirstInputMixin:
    @property
    def first_input_type(self) -> type:
        if (previous := getattr(self, "previous", None)) is not None:
            if (first_input_type := getattr(previous, "first_input_type", None)) is not None:
                return first_input_type
            else:
                raise TypeError(
                    f"Cannot determine first input type because previous {type(previous)} does not have it",
                )  # pragma: no cover
        else:
            from chorey.step import Step
            from chorey.utils import get_input_output_type_from_step

            if isinstance(self, Step):
                input_type, _ = get_input_output_type_from_step(self)
                return input_type
            else:
                raise TypeError(
                    f"Cannot determine first input type for {type(self)} without previous step",
                )  # pragma: no cover



from dataclasses import dataclass
from typing import TypeVar

from pype import PypelineContext, step
from pype.mermaid import mermaid


@dataclass(frozen=True)
class Context:
    force_full_analysis: bool
    db_session: str  # Placeholder for an actual DB session object


@dataclass
class Document:
    id: int
    content: str


@dataclass
class Summary:
    document_id: int
    summary_text: str


@dataclass
class FullAnalysis:
    document_id: int
    word_count: int
    sentiment: str


# helper types to make the signatures cleaner
T = TypeVar("T")
Input = PypelineContext[T, Context]
Output = Input


async def fetch_document(input: Input[int]) -> Output[Document]:
    # unpack the PypelineContext to work with its parts
    doc_id, ctx = input.parts

    # simulate fetching a document from a database or API

    # use .with_data() to keep the context while changing the data
    return input.with_data(Document(id=doc_id, content="This is a sample document content."))


async def summarize_document(input: Input[Document]) -> Output[Summary]:
    # unpack the PypelineContext to work with its parts
    doc, ctx = input.parts

    # simulate summarizing the document

    return input.with_data(Summary(document_id=doc.id, summary_text="This is a summary."))


async def analyze_document(input: Input[Document]) -> Output[FullAnalysis]:
    doc, ctx = input.parts

    # simulate performing a full analysis of the document

    return input.with_data(FullAnalysis(document_id=doc.id, word_count=len(doc.content.split()), sentiment="Positive"))


async def store_processing_result(input: Input[Summary] | Input[FullAnalysis]) -> None:
    result, ctx = input.parts

    # simulate storing the result
    print(f"Storing result: {result}")
    print(f"Type of result: {type(result)}")


def selector(input: Input[Document]) -> int:
    doc, ctx = input.parts

    if ctx.force_full_analysis:
        return 1  # force full analysis if the context says so

    # same as before
    return 0 if len(doc.content) < 50 else 1


pipeline = (
    step(fetch_document)
    .route(step(summarize_document), step(analyze_document), selector=selector, decision_label="Document Length Check")
    .next(store_processing_result)
)

if __name__ == "__main__":
    print(mermaid(pipeline))



from typing import Awaitable, Callable, Protocol, TypeVar

FirstInput = TypeVar("FirstInput", contravariant=True)
Input = TypeVar("Input", contravariant=True)
Output = TypeVar("Output", covariant=True)


StepFunc = Callable[[Input], Awaitable[Output]]


class Feeder(Protocol[FirstInput, Input, Output]):
    async def __call__(self, input: Input) -> Output:
        """
        Processes an input of type `Input` and produces an output of type `Output`.
        """

        ...

    async def feed(self, input: FirstInput) -> Output:
        """
        Feeds a chain of Feeders starting from an initial input of type `FirstInput`
        and produces an output of type `Output`.
        """

        ...



from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from chorey.step import Step


class PipelineError(Exception):
    step_name: str
    input: Any
    original_exception: Exception

    def __init__(self, *, step: "Step", input: Any, original_exception: Exception) -> None:
        self.step_name = step.name
        self.input = input
        self.original_exception = original_exception

        message = (
            f"Pipeline execution failed at step '{self.step_name}'",
            f"Input that caused the error: {input!r}",
            f"Original error: {type(original_exception).__name__}: {original_exception}",
        )

        super().__init__("\n".join(message))


class UnreachableCodeReachedError(Exception):
    """
    Exception raised when the code reaches a point that should be logically unreachable.
    """

    def __init__(self) -> None:
        super().__init__("Unreachable code reached.")



from dataclasses import dataclass
from typing import Generic, TypeVar

T = TypeVar("T")
R = TypeVar("R")
Context = TypeVar("Context")


@dataclass
class ChoreyContext(Generic[T, Context]):
    """
    Wraps data with additional context information.
    Recommended for use as input and output types for nodes in a pype pipeline,
    as when visualizing the pipeline, the data type is shown instead of this wrapper or
    the context.

    Example:
    ```python
    from dataclasses import dataclass
    from typing import TypeVar
    from chorey.context import ChoreyContext
    from chorey.node import node
    from chorey.display import print_table_from_node

    @dataclass
    class Context:
        id: str

    T = TypeVar("T")
    Input = ChoreyContext[T, Context]
    Output = Input

    async def first(input: Input[int]) -> Output[str]:
        # access context information
        print(input.context.id)
        return input.with_data("example" * input.data)

    async def second(input: Input[str]) -> Output[float]:
        ...

    pipeline = node(first).next(second)

    pipeline.feed(Input(data=42, context=Context(id="example")))

    # displays input and output types without the ChoreyContext wrapper
    # e.g. int -> str -> float
    print(print_table_from_node(pipeline))
    ```
    """

    data: T
    context: Context

    def with_data(self, new_data: R) -> "ChoreyContext[R, Context]":
        """
        Creates a new `ChoreyContext` with the same context but different data.
        """

        return ChoreyContext(data=new_data, context=self.context)

    @property
    def parts(self) -> tuple[T, Context]:
        return (self.data, self.context)



import asyncio
from dataclasses import dataclass, field
from types import EllipsisType
from typing import Any, Awaitable, Callable, Final, Generic, TypeVar

from pydantic import BaseModel
from pydantic import ConfigDict as ModelConfig
from pydantic import Field

from chorey._types import Feeder, FirstInput, Input, Output, StepFunc
from chorey.error import PipelineError, UnreachableCodeReachedError
from chorey.mixins import BranchableMixin, GetFirstInputMixin, NextableMixin, RoutableMixin
from chorey.utils import get_step_func_name

R = TypeVar("R")

OnFailureFunc = Callable[[Input, Exception], None | Awaitable[None]]
"""
Callback function to be called when the step fails.
Takes the input that caused the failure and the exception as arguments.
Must return None as it is only for side effects (e.g., logging).
"""


class FailureConfig(BaseModel, Generic[Input]):
    max_attempts: int = Field(default=1, ge=1)
    delay_seconds: float = Field(default=0.0, ge=0.0)
    on_failure: OnFailureFunc[Input] | None = None

    model_config = ModelConfig(frozen=True)

    def copy_with(
        self,
        *,
        max_attempts: int | None = None,
        delay_seconds: float | None = None,
        on_failure: OnFailureFunc[Input] | None | EllipsisType = ...,
    ) -> "FailureConfig":
        return FailureConfig(
            max_attempts=max_attempts if max_attempts is not None else self.max_attempts,
            delay_seconds=delay_seconds if delay_seconds is not None else self.delay_seconds,
            on_failure=self.on_failure if on_failure is ... else on_failure,
        )


@dataclass(frozen=True)
class Step(
    RoutableMixin[FirstInput, Input, Output],
    BranchableMixin[FirstInput, Input, Output],
    NextableMixin[FirstInput, Input, Output],
    GetFirstInputMixin,
    Feeder[FirstInput, Input, Output],
):
    """
    Represents a step in a processing pipeline that mutates the data it receives.
    Each step is linked to its previous one to allow chaining of operations.

    Note:
        The reason why backward linkage is used is to ensure that the first input type is preserved
        throughout the chain of steps.
    """

    func: Final[StepFunc[Input, Output]]
    """ The function that the step will execute. """

    name: Final[str]
    """ The name of the step, either given or derived from the function name. """

    description: Final[str | None] = None
    """
    An optional description of what the step does. Will be displayed in the graph output.
    """

    failure_config: FailureConfig | None = field(default=None)
    """
    Configuration for handling failures in the step.
    """

    previous: "Feeder[FirstInput, Input, Any] | None" = field(default=None)
    """
    A reference to the previous step in the chain. None if this is the first step.
    """

    async def __call__(self, input: Input) -> Output:
        """
        Calls the step's function with the given input and returns the output.
        """

        failure_config = self.failure_config or FailureConfig()

        for attempt in range(1, failure_config.max_attempts + 1):
            try:
                return await self.func(input)
            except Exception as e:
                if attempt == failure_config.max_attempts:
                    if failure_config.on_failure is not None:
                        if asyncio.iscoroutinefunction(failure_config.on_failure):
                            await failure_config.on_failure(input, e)
                        else:
                            failure_config.on_failure(input, e)

                    raise PipelineError(step=self, original_exception=e, input=input) from e
                else:
                    if failure_config.delay_seconds > 0:
                        await asyncio.sleep(failure_config.delay_seconds)

        raise UnreachableCodeReachedError()

    async def feed(self, input: FirstInput) -> Output:
        """
        Feeds the input through the chain of steps, starting from the first step to the current one.
        """

        if self.previous is not None:
            intermediate = await self.previous.feed(input)
            return await self(intermediate)
        else:
            return await self(input)  # type: ignore

    def copy_with(
        self,
        *,
        previous: "Feeder[FirstInput, Input, Any] | None | EllipsisType" = ...,
        failure_config: FailureConfig | None | EllipsisType = ...,
    ) -> "Step[FirstInput, Input, Output]":
        return Step(
            func=self.func,
            name=self.name,
            description=self.description,
            failure_config=self.failure_config if failure_config is ... else failure_config,
            previous=self.previous if previous is ... else previous,
        )

    def retry(
        self,
        max_attempts: int = 3,
        delay_seconds: float = 1.0,
    ) -> "Step[FirstInput, Input, Output]":
        return self.copy_with(
            failure_config=(self.failure_config or FailureConfig()).copy_with(
                max_attempts=max_attempts,
                delay_seconds=delay_seconds,
            )
        )

    def on_failure_do(
        self,
        func: OnFailureFunc,
    ) -> "Step[FirstInput, Input, Output]":
        """
        Sets a callback function to be called if the step fails.
        The function takes the input that caused the failure and the exception as arguments.
        """

        return self.copy_with(
            failure_config=(self.failure_config or FailureConfig()).copy_with(
                on_failure=func,
            )
        )


def step(
    func: StepFunc[Input, Output],
    /,
    *,
    description: str | None = None,
    name: str | None = None,
) -> Step[Input, Input, Output]:
    """
    Sets up the initial step in the processing pipeline.
    """

    return Step(func, description=description, name=name or get_step_func_name(func))



import asyncio
from dataclasses import dataclass, field
from types import EllipsisType
from typing import TYPE_CHECKING, Any, Generic, TypeVar, TypeVarTuple

from chorey._types import Feeder, FirstInput, Input, StepFunc
from chorey.mixins import GetFirstInputMixin
from chorey.utils import get_step_func_name

if TYPE_CHECKING:
    from chorey.step import Step

OrderedOutcomes = TypeVarTuple("OrderedOutcomes")
R = TypeVar("R")


@dataclass(frozen=True)
class Branch(
    Generic[FirstInput, Input, *OrderedOutcomes],
    GetFirstInputMixin,
    Feeder[FirstInput, Input, tuple[*OrderedOutcomes]],
):
    steps: list["Step[Input, Any, Any]"] = field(default_factory=list)

    previous: Feeder[FirstInput, Input, Any] | None = field(default=None)

    async def __call__(self, input: Input) -> tuple[*OrderedOutcomes]:
        results = await asyncio.gather(*(step.feed(input) for step in self.steps))
        return tuple(results)  # type: ignore

    async def feed(
        self,
        input: FirstInput,
    ) -> tuple[*OrderedOutcomes]:
        if self.previous is not None:
            intermediate = await self.previous.feed(input)
            return await self(intermediate)
        else:
            return await self(input)  # type: ignore

    def copy_with(
        self,
        *,
        steps: list["Step[Input, Any, Any]"] | None = None,
        previous: Feeder[FirstInput, Input, Any] | None | EllipsisType = ...,
    ) -> "Branch[FirstInput, Input, *OrderedOutcomes]":
        return Branch(
            steps=self.steps if steps is None else steps,
            previous=self.previous if previous is ... else previous,
        )

    def merge(
        self,
        func: StepFunc[tuple[*OrderedOutcomes], R],
        /,
        *,
        name: str | None = None,
        description: str | None = None,
    ) -> "Step[FirstInput, tuple[*OrderedOutcomes], R]":
        from chorey.step import Step

        return Step(
            func=func,
            name=name or get_step_func_name(func),
            description=description,
            previous=self,  # type: ignore
        )



import asyncio
from typing import Annotated

from chorey.utils import MissingFeatureError, import_pipeline, parse_cli_input

try:
    import typer
except ImportError:
    raise MissingFeatureError("cli")


app = typer.Typer(
    name="chorey",
    help="A CLI tool to run and visualize Chorey pipelines defined in Python modules.",
)


@app.command(
    short_help="Run a Chorey pipeline defined in a Python module.",
)
def run(
    pipeline_path: str = typer.Argument(
        ...,
        help="Path to the pipeline module, e.g., 'my_pipeline.py' or 'my_module:my_pipeline'",
    ),
    input: Annotated[
        str | None,
        typer.Option(
            "--input",
            "-i",
            help="Input value to feed into the pipeline. Can be primitive or JSON-encoded.",
        ),
    ] = None,
):
    pipeline = import_pipeline(pipeline_path)
    first_input_type = pipeline.first_input_type
    parsed_input = parse_cli_input(input, first_input_type)

    result = asyncio.run(pipeline.feed(parsed_input))
    print(result)


@app.command(
    short_help="Generate a Mermaid diagram of the pipeline structure.",
)
def visualize(
    pipeline_path: str = typer.Argument(
        ...,
        help="Path to the pipeline module, e.g., 'my_pipeline.py' or 'my_module:my_pipeline'",
    ),
):
    from chorey import mermaid

    pipeline = import_pipeline(pipeline_path)
    print(mermaid(pipeline))


if __name__ == "__main__":
    app()



from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Generator, Iterator, Protocol, TypeGuard

from chorey.utils import MissingFeatureError, get_input_output_type_from_step, get_type_name_for_display

try:
    from uuid import uuid4
except ImportError as e:
    raise MissingFeatureError("mermaid") from e

if TYPE_CHECKING:
    from chorey import Step


class _UnexpectedTypeError(TypeError):
    def __init__(self, ty: type | object) -> None:
        type_ = ty if isinstance(ty, type) else type(ty)

        super().__init__(f"Unexpected type: {type_}")


def _new_id() -> str:
    """
    Generates a new unique identifier for nodes in the mermaid diagram.
    They will always start with "id-" to ensure they do not start with a digit,
    as mermaid does not support node IDs that start with them.
    """

    return f"id-{uuid4().hex}"


@dataclass
class _Edge:
    """
    Represents an edge in the mermaid diagram, connecting nodes.
    """

    id: str
    type: str


class _MermaidGenerator(Protocol):
    def generate(self, *edges: _Edge) -> Iterator[str]:
        """Generate mermaid edge definitions from given from-edges."""
        ...

    @property
    def out_edges(self) -> Iterator[_Edge]:
        """Get the outgoing edges from this node."""
        ...


class _HasIdNameDescription(Protocol):
    id: str
    name: str
    description: str | None


class _MermaidNode(_HasIdNameDescription, _MermaidGenerator, Protocol):
    """
    Represents a node datastructure in the mermaid diagram.
    """

    pass


@dataclass
class _Input(_MermaidNode):
    """
    Input node for the mermaid diagram. Only one is used per diagram as the starting point.
    """

    out_type: str
    id: str = field(default="input", init=False)
    name: str = field(default="Input", init=False)
    description: str | None = field(default=None, init=False)

    def generate(self, *edges: _Edge) -> Generator[str, None, None]:
        yield f'    {self.id}["{self.name}"]'

    @property
    def out_edges(self) -> Iterator[_Edge]:
        yield _Edge(id=self.id, type=self.out_type)


@dataclass
class _Output(_MermaidNode):
    """
    Output node for the mermaid diagram. Only one is used per diagram as the ending point.
    """

    id: str = field(default="output", init=False)
    name: str = field(default="Output", init=False)
    description: str | None = field(default=None, init=False)

    def generate(self, *edges: _Edge) -> Iterator[str]:
        for edge in edges:
            yield f'    {edge.id} -->|"{edge.type}"| {self.id}["{self.name}"]'

    @property
    def out_edges(self) -> Iterator[_Edge]:
        yield from ()


@dataclass
class _Step(_MermaidNode):
    """
    `Step`'s representation for the mermaid diagram.
    """

    name: str
    description: str | None
    input_type: str
    output_type: str
    id: str = field(default_factory=_new_id)

    def generate(self, *edges: _Edge) -> Iterator[str]:
        label = f"<b>{self.name}</>"
        if self.description is not None:
            label += f"</br><sub>{self.description}</sub>"

        for edge in edges:
            yield f'    {edge.id} -->|"{edge.type}"| {self.id}["{label}"]'

    @property
    def out_edges(self) -> Iterator[_Edge]:
        yield _Edge(id=self.id, type=self.output_type)


@dataclass
class _Branch(_MermaidNode):
    """
    `Branch`'s representation for the mermaid diagram.
    """

    branches: list[list[_MermaidNode]]
    """
    A list of branches, where each branch is a list of `_MermaidNode` instances.
    """

    # these fields are not used for branches, probably should be refactored later
    id: str = field(default_factory=_new_id, init=False)
    name: str = field(default="<BRANCH>", init=False)
    description: str | None = field(default=None, init=False)

    def generate(self, *edges: _Edge) -> Generator[str, None, None]:
        for branch in self.branches:
            if not branch:
                continue

            # the branch will be from start-to-end at first, but we actually
            # need it to be from end-to-start to generate the edges correctly
            branch = list(reversed(branch))

            # generate the edges up until the last (actually first) step in the branch
            for current, next in zip(branch, branch[1:]):
                yield from current.generate(*next.out_edges)

            # finally, the last step in the branch (actually the first)
            # will connect the previous edges to the start of this branch
            yield from branch[-1].generate(*edges)

    @property
    def out_edges(self) -> Iterator[_Edge]:
        # we only need to yield the out_edges of the last step in each branch
        # because those are the ones that will connect to the next step after the branch
        for branch in self.branches:
            if not branch:
                continue
            yield from branch[-1].out_edges

    @property
    def input_type(self) -> str:
        # will never happen, hopefully
        if not self.branches:
            return "None"

        # since all paths in a branch must have the same input type at first,
        # we can just take the first branch and the first generator in it
        # and return its input type
        first_generator = self.branches[0][0]

        if _ensure_generator_type(first_generator):
            return first_generator.input_type

        raise _UnexpectedTypeError(first_generator)


@dataclass
class _Route(_MermaidNode):
    """
    `Branch`'s representation for the mermaid diagram.
    """

    branches: list[list[_MermaidNode]]
    """
    A list of branches the router chooses rom, where each branch is a list of `_MermaidNode` instances.
    """

    decision_label: str | None

    # just like _Branch, these fields are also not used
    id: str = field(default_factory=_new_id, init=False)
    name: str = field(default="<ROUTE>", init=False)
    description: str | None = field(default=None, init=False)

    def generate(self, *edges: _Edge) -> Generator[str, None, None]:
        label = self.decision_label or ""
        yield f'    {self.id}{{"{label}"}}'

        for edge in edges:
            yield f'    {edge.id} -->|"{edge.type}"| {self.id}'

        for i, branch in enumerate(self.branches):
            if not branch:
                continue

            branch = list(reversed(branch))

            for current, next in zip(branch, branch[1:]):
                yield from current.generate(*next.out_edges)

            start_of_choice = branch[-1]

            if _ensure_generator_type(start_of_choice):
                condition_label = f"Choice {i}"
                yield from start_of_choice.generate(
                    _Edge(
                        id=self.id,
                        type=f"{condition_label} ({start_of_choice.input_type})",
                    )
                )
            else:
                raise _UnexpectedTypeError(start_of_choice)

    @property
    def out_edges(self) -> Iterator[_Edge]:
        for branch in self.branches:
            if not branch:
                continue
            yield from branch[-1].out_edges

    @property
    def input_type(self) -> str:
        if not self.branches:
            return "None"

        first_generator = self.branches[0][0]

        if _ensure_generator_type(first_generator):
            return first_generator.input_type

        raise _UnexpectedTypeError(first_generator)


def _ensure_generator_type(g: _MermaidGenerator) -> TypeGuard[_Step | _Branch | _Route]:
    """
    Currently only `_Step` and `_Branch` are valid types for generators in the mermaid diagram.
    But maybe in the future there will be more types, that is why I generalized almost everything
    in this file.
    """

    return isinstance(g, (_Step, _Branch, _Route))


def mermaid(step: "Step") -> str:
    """
    Generate a mermaid diagram from a given pipeline.
    The generated string can be used in mermaid live editor:
    https://mermaid.live/
    """

    from chorey import Branch, Route, Step

    # the lines of the mermaid diagram
    lines: list[str] = []
    lines.append("flowchart TD")

    def visit(n: Step | Branch | Route) -> Iterator[_MermaidNode]:
        """
        Recursively turns a `Step` or `Branch` into their mermaid representation.
        """

        if isinstance(n, Step):
            input_type, output_type = get_input_output_type_from_step(n)

            yield _Step(
                name=n.name,
                description=n.description,
                input_type=get_type_name_for_display(input_type),
                output_type=get_type_name_for_display(output_type),
            )
        elif isinstance(n, Branch):
            branches: list[list[_MermaidNode]] = []
            for branch in n.steps:
                items: list[_MermaidNode] = []
                items.extend(collect(branch))
                branches.append(items)
            yield _Branch(branches=branches)
        elif isinstance(n, Route):
            branches: list[list[_MermaidNode]] = []
            for branch in n.choices:
                items: list[_MermaidNode] = []
                items.extend(collect(branch))
                branches.append(items)
            yield _Route(branches=branches, decision_label=n.decision_label)
        else:
            raise _UnexpectedTypeError(n)

    def collect(n: Step | Branch | Route) -> Iterator[_MermaidNode]:
        """
        Flattens the pipeline into a series of mermaid nodes.
        """

        if n.previous is None:
            yield from visit(n)
        elif isinstance(n.previous, (Step, Branch, Route)):
            yield from collect(n.previous)
            yield from visit(n)
        else:
            raise _UnexpectedTypeError(n.previous)

    items: list[_MermaidNode] = []

    pipeline_items = list(collect(step))

    if (item := pipeline_items[0]) and _ensure_generator_type(item):
        items.append(_Input(out_type=item.input_type))
    else:
        raise _UnexpectedTypeError(item)

    items.extend(pipeline_items)
    items.append(_Output())

    # we need to generate the edges from the end to the start
    items.reverse()

    # generate the edges between all nodes
    for current, next in zip(items, items[1:]):
        lines.extend(current.generate(*next.out_edges))

    # finally, the input node
    lines.extend(items[-1].generate(*items[-2].out_edges))

    return "\n".join(lines)



from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Callable, TypeVar

from chorey._types import Feeder, FirstInput, Input, Output
from chorey.mixins import BranchableMixin, GetFirstInputMixin, NextableMixin, RoutableMixin

if TYPE_CHECKING:
    from chorey.step import Step

T = TypeVar("T")


@dataclass(frozen=True, kw_only=True)
class Route(
    NextableMixin[FirstInput, Input, Output],
    BranchableMixin[FirstInput, Input, Output],
    RoutableMixin[FirstInput, Input, Output],
    GetFirstInputMixin,
    Feeder[FirstInput, Input, Output],
):
    choices: list["Step[Input, Any, Output]"] = field(default_factory=list)
    selector: Callable[[Input], int] = field(default=lambda _: 0)

    decision_label: str | None = field(default=None)
    """
    The label to use for the decision point in the mermaid diagram.
    If None, empty diamond will be used.
    """

    previous: Feeder[FirstInput, Input, Any] | None = field(default=None)

    async def __call__(self, input: Input) -> Output:
        index = self.selector(input)
        if index < 0 or index >= len(self.choices):
            raise IndexError(f"Selector returned out-of-bounds index {index} for choices of length {len(self.choices)}")

        step = self.choices[index]
        return await step.feed(input)

    async def feed(
        self,
        input: FirstInput,
    ) -> Output:
        if self.previous is not None:
            intermediate = await self.previous.feed(input)
            return await self(intermediate)
        else:
            return await self(input)  # type: ignore



from types import NoneType
from typing import TYPE_CHECKING, Any, TypeVar, get_args, get_origin

from pydantic import BaseModel

from chorey._types import StepFunc
from chorey.context import ChoreyContext

if TYPE_CHECKING:
    from chorey.step import Step

Input = TypeVar("Input")
Output = TypeVar("Output")


def get_step_func_name(func: "StepFunc") -> str:
    return getattr(func, "__name__", repr(func))


def get_input_output_type_from_step(
    step: "Step[Input, Output, Any]",
) -> tuple[type[Input], type[Output]]:
    from inspect import signature

    func_name = get_step_func_name(step.func)
    sig = signature(step.func)

    input_parameter = tuple(sig.parameters.values())[0] if sig.parameters else None
    if input_parameter is None or input_parameter.annotation is sig.empty:
        raise TypeError(f"Function {func_name} must have a type-annotated first parameter")

    output_type = sig.return_annotation
    if output_type is sig.empty:
        raise TypeError(f"Function {func_name} must have a return type annotation")

    return input_parameter.annotation, output_type


class MissingFeatureError(Exception):
    """
    Exception raised when a required feature is missing.
    """

    def __init__(self, feature: str) -> None:
        super().__init__(f"Missing required feature for pype: {feature}")  # pragma: no cover


def get_type_name_for_display(
    ty: type,
    *,
    unwrap_pypeline_context: bool = True,
) -> str:
    """
    Get a user-friendly name for a type, suitable for display in the pipeline visualization.
    """

    # unwrap ChoreyContext's data type for display purposes, as context does not matter
    # in the visualization
    if unwrap_pypeline_context and get_origin(ty) is ChoreyContext:
        ty, _ = get_args(ty)

    if args := get_args(ty):
        origin = get_origin(ty)
        if origin is not None:
            arg_names = ", ".join(get_type_name_for_display(arg) for arg in args)
            return f"{get_type_name_for_display(origin)}[{arg_names}]"
    elif ty is NoneType:
        return "None"

    if hasattr(ty, "__name__"):
        return ty.__name__

    return str(ty)  # pragma: no cover


class _NotExisting:
    pass


class ImportPipelineError(Exception):
    pass


class InvalidPipelinePathError(ImportPipelineError):
    def __init__(self, path: str) -> None:
        super().__init__(f"Invalid pipeline path: '{path}'. Must be in the format 'module.submodule:variable_name'")


class PipelineModuleDoesNotExistError(ImportPipelineError):
    def __init__(self, module: str) -> None:
        super().__init__(f"Could not find module to import pipeline from: '{module}'")


class PipelineDoesNotExistError(ImportPipelineError):
    def __init__(self, path: str) -> None:
        super().__init__(f"Could not find pipeline variable in module: '{path}'")


def import_pipeline(path: str) -> "Step":
    from chorey.step import Step

    if not path.count(":") == 1:
        raise InvalidPipelinePathError(path)

    module_path, var_name = path.rsplit(":", 1)

    try:
        module = __import__(module_path, fromlist=[var_name])
    except ModuleNotFoundError as e:
        raise PipelineModuleDoesNotExistError(module_path) from e

    var = getattr(module, var_name, _NotExisting)
    if var is _NotExisting:
        raise PipelineDoesNotExistError(var_name)

    if not isinstance(var, Step):
        raise TypeError(f"'{var_name}' is not a Step instance")

    return var


def parse_cli_input(value: str | None, to_type: type) -> Any:
    if value is None:
        if to_type is type(None):
            return None
        else:
            to_type_name = get_type_name_for_display(to_type, unwrap_pypeline_context=False)
            raise TypeError(f"Input of type '{to_type_name}' is required but none was provided")

    if isinstance(value, to_type):
        return value

    from json import JSONDecodeError, loads

    if BaseModel in to_type.__mro__:
        try:
            return to_type(**loads(value))
        except JSONDecodeError:
            raise TypeError(f"Cannot parse input into {to_type}: expected JSON object, got {type(value)}")

    if to_type not in {str, int, float, bool}:
        try:
            value = loads(value)
        except JSONDecodeError:
            value = value

    return to_type(value)
